from copy import deepcopy

from openfisca_core.taxbenefitsystems import TaxBenefitSystem
from openfisca_core.tools import assert_near

from openfisca_france.entities import entities, Individu, Famille, Menage
from openfisca_france.model.base import *  # noqa analysis:ignore


class af(Variable):
    value_type = float
    entity = Famille
    definition_period = MONTH
    set_input = set_input_divide_by_period


class salaire(Variable):
    value_type = float
    entity = Individu
    definition_period = MONTH
    set_input = set_input_divide_by_period


class age(Variable):
    value_type = int
    entity = Individu
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class autonomie_financiere(Variable):
    value_type = bool
    entity = Individu
    definition_period = MONTH
    set_input = set_input_divide_by_period


class depcom(Variable):
    value_type = str
    max_length = 5
    entity = Menage
    label = '''Code INSEE "depcom" de la commune de résidence de la famille'''
    definition_period = ETERNITY

# This tests are more about core than france, but we need france entities to run some of them.
# We use a dummy TBS to run the tests faster


class DummyTaxBenefitSystem(TaxBenefitSystem):
    def __init__(self):
        TaxBenefitSystem.__init__(self, entities)
        self.add_variables(af, salaire, age, autonomie_financiere, depcom)


tax_benefit_system = DummyTaxBenefitSystem()

TEST_CASE = {
    'individus': {'ind0': {}, 'ind1': {}, 'ind2': {}, 'ind3': {}, 'ind4': {}, 'ind5': {}},
    'familles': {
        'f1': {'enfants': ['ind2', 'ind3'], 'parents': ['ind0', 'ind1']},
        'f2': {'enfants': ['ind5'], 'parents': ['ind4']}
        },
    'foyers_fiscaux': {
        'ff1': {'declarants': ['ind0', 'ind1'], 'personnes_a_charge': ['ind2', 'ind3']},
        'ff2': {'personnes_a_charge': ['ind5'], 'declarants': ['ind4']}
        },
    'menages': {
        'm1': {'conjoint': 'ind1', 'enfants': ['ind2', 'ind3'], 'personne_de_reference': 'ind0'},
        'm2': {'conjoint': [], 'enfants': ['ind5'], 'personne_de_reference': 'ind4'},
        },
    }

TEST_CASE_AGES = deepcopy(TEST_CASE)
AGES = [40, 37, 7, 9, 54, 20]

for (individu, age_) in zip(TEST_CASE_AGES['individus'].values(), AGES):
    individu['age'] = age_

reference_period = 2013


def new_simulation(test_case):
    test_case['period'] = reference_period
    return tax_benefit_system.new_scenario().init_from_dict(test_case).new_simulation()


def test_transpose():
    test_case = deepcopy(TEST_CASE)
    test_case['familles']['f1']['af'] = 20000
    test_case['familles']['f2']['af'] = 10000
    test_case['foyers_fiscaux'] = {
        'ff1': TEST_CASE['foyers_fiscaux']['ff1'],
        'ff2': {'declarants': ['ind4']},
        'ff3': {'declarants': ['ind5']}
        }

    simulation = new_simulation(test_case)
    foyer_fiscal = simulation.foyer_fiscal

    af_foyer_fiscal = foyer_fiscal.first_person.famille('af', period = reference_period, options = [ADD])

    assert_near(af_foyer_fiscal, [20000, 10000, 10000])


def test_transpose_string():
    test_case = deepcopy(TEST_CASE)
    test_case['menages']['m1']['depcom'] = '93400'
    test_case['menages']['m2']['depcom'] = '89300'

    simulation = new_simulation(test_case)
    famille = simulation.famille

    depcom_famille = famille.first_person.menage('depcom', period = reference_period)

    assert (depcom_famille == [b'93400', b'89300']).all()


def test_value_from_person():
    test_case = deepcopy(TEST_CASE_AGES)
    simulation = new_simulation(test_case)

    foyer_fiscal = simulation.foyer_fiscal

    age_conjoint = foyer_fiscal.conjoint('age', period='2013-01')
    assert_near(age_conjoint, [37, 0])


def test_combination_projections():
    test_case = deepcopy(TEST_CASE_AGES)
    simulation = new_simulation(test_case)

    individu = simulation.persons

    age_parent1 = individu.famille.demandeur('age', period='2013-01')

    assert_near(age_parent1, [40, 40, 40, 40, 54, 54])


def test_complex_chain_2():
    test_case = {
        'individus': {'ind0': {'age': 30}, 'ind1': {'age': 31}, 'ind2': {'age': 32}, 'ind3': {'age': 33}},
        'familles': {
            'f1': {'parents': ['ind0', 'ind1']},
            'f2': {'parents': ['ind2']},
            'f3': {'parents': ['ind3']},
            },
        'foyers_fiscaux': {
            'ff1': {'declarants': ['ind0', 'ind1']},
            'ff2': {'declarants': ['ind2', 'ind3']},
            },
        'menages': {
            'm1': {'personne_de_reference': 'ind0'},
            'm2': {'personne_de_reference': 'ind1', 'autres': 'ind2'},
            'm4': {'personne_de_reference': 'ind3'},
            },
        }

    simulation = new_simulation(test_case)

    assert_near(simulation.famille.demandeur.menage.personne_de_reference('age', period='2013-01'), [30, 31, 33])
    assert_near(simulation.famille.conjoint.menage.personne_de_reference('age', period='2013-01'), [31, 0, 0])
    assert_near(simulation.famille.demandeur.foyer_fiscal.declarant_principal('age', period='2013-01'), [30, 32, 32])
    assert_near(simulation.foyer_fiscal.conjoint.famille.demandeur('age', period='2013-01'), [30, 33])



from openfisca_france.model.base import *

from openfisca_france.model.prestations.education import TypesScolarite


class aide_formation_gen_eligibilite(Variable):
    value_type = bool
    entity = Individu
    reference = 'https://www.legifrance.gouv.fr/loda/id/JORFTEXT000034429379/'
    label = "Éligibilité à l'aide accordée aux personnes inscrites dans une formation labellisée par la Grande Ecole du numérique"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period

    def formula_2017_04_14(individu, period):
        '''
        Les conditions de non-cumul spécifiques à la formation ne sont pas modélisées car aucune n'est modélisée à ce jour dans OpenFisca et qu'un grand nombre sont arbitraires ou à l'échelon infra-national :
        - L'aide n'est pas cumulable avec une aide spécifique versée par le ministre chargé de l'enseignement supérieur ou les établissements publics qui en relèvent.
        - Sont exclues du bénéfice de l'aide les personnes inscrites à Pôle Emploi comme demandeurs d'emploi qui perçoivent une aide à l'insertion ou une aide à la formation professionnelle ainsi que […] les personnes en congé individuel de formation.
        - Sont également exclues du bénéfice de l'aide les personnes percevant une aide du ministère chargé de l'emploi ou d'un conseil régional versée au titre de la formation professionnelle ou de l'insertion professionnelle.
        '''
        eligibilite_formation = individu('scolarite', period) == TypesScolarite.grande_ecole_du_numerique
        eligibilite_nationalite = individu('bourse_criteres_sociaux_eligibilite_nationalite', period)

        non_cumul = individu('alternant', period) + (
            individu.famille('rsa', period) * not_(individu.famille('rsa_majore_eligibilite', period))
            )

        return eligibilite_formation * eligibilite_nationalite * not_(non_cumul)


class aide_formation_gen(Variable):
    value_type = float
    entity = Individu
    reference = 'https://www.legifrance.gouv.fr/loda/id/JORFTEXT000034429379/'
    label = "Montant de l'aide accordée aux personnes inscrites dans une formation labellisée par la Grande Ecole du numérique"
    definition_period = MONTH
    set_input = set_input_divide_by_period

    def formula_2017_04_14(individu, period, parameters):
        points_de_charge = individu('bourse_criteres_sociaux_points_de_charge', period)
        baremes = parameters(period).prestations_sociales.aides_jeunes.bourses.bourses_enseignement_superieur.criteres_sociaux.plafond_ressources
        plafond_echelon_0bis = baremes.echelon_0bis.calc(points_de_charge)
        plafond_echelon_1 = baremes.echelon_1.calc(points_de_charge)
        plafond_echelon_2 = baremes.echelon_2.calc(points_de_charge)
        plafond_echelon_3 = baremes.echelon_3.calc(points_de_charge)
        plafond_echelon_4 = baremes.echelon_4.calc(points_de_charge)
        plafond_echelon_5 = baremes.echelon_5.calc(points_de_charge)
        plafond_echelon_6 = baremes.echelon_6.calc(points_de_charge)
        plafond_echelon_7 = baremes.echelon_7.calc(points_de_charge)

        base_ressources = individu('bourse_criteres_sociaux_base_ressources', period)
        echelon = select(
            [
                base_ressources <= plafond_echelon_7,
                base_ressources <= plafond_echelon_6,
                base_ressources <= plafond_echelon_5,
                base_ressources <= plafond_echelon_4,
                base_ressources <= plafond_echelon_3,
                base_ressources <= plafond_echelon_2,
                base_ressources <= plafond_echelon_1,
                base_ressources <= plafond_echelon_0bis,
                ],
            [7, 6, 5, 4, 3, 2, 1, 0], default=-1)

        montants = parameters(period).prestations_sociales.aides_jeunes.bourses.bourses_enseignement_superieur.criteres_sociaux.montants

        return montants.calc(echelon) * individu('aide_formation_gen_eligibilite', period)



from openfisca_core.entities import build_entity

Famille = build_entity(
    key = 'famille',
    plural = 'familles',
    label = 'Famille',
    roles = [
        {
            'key': 'parent',
            'plural': 'parents',
            'label': 'Parents',
            'subroles': ['demandeur', 'conjoint']
            },
        {
            'key': 'enfant',
            'plural': 'enfants',
            'label': 'Enfants'
            }
        ]
    )

Individu = build_entity(
    key = 'individu',
    plural = 'individus',
    label = 'Individu',
    is_person = True
    )

FoyerFiscal = build_entity(
    key = 'foyer_fiscal',
    plural = 'foyers_fiscaux',
    label = 'Déclaration d’impôts',
    doc = '''
    Le foyer fiscal désigne l'ensemble des personnes inscrites sur une même déclaration de revenus.
    Il peut y avoir plusieurs foyers fiscaux dans un seul ménage : par exemple, un couple non marié où chacun remplit
    sa propre déclaration de revenus compte pour deux foyers fiscaux.
    Voir https://www.insee.fr/fr/metadonnees/definition/c1735.
    ''',
    roles = [
        {
            'key': 'declarant',
            'plural': 'declarants',
            'label': 'Déclarants',
            'subroles': ['declarant_principal', 'conjoint'],
            },
        {
            'key': 'personne_a_charge',
            'plural': 'personnes_a_charge',
            'label': 'Personnes à charge'
            },
        ]
    )

Menage = build_entity(
    key = 'menage',
    plural = 'menages',
    label = 'Logement principal',
    doc = '''
    Un ménage, au sens statistique du terme, désigne l'ensemble des occupants d'un même logement sans que ces personnes
    soient nécessairement unies par des liens de parenté (en cas de cohabitation, par exemple).
    Un ménage peut être composé d'une seule personne.
    Le niveau de vie ainsi que la pauvreté sont calculés au niveau d'un ménage.
    Voir https://www.insee.fr/fr/metadonnees/definition/c1879.
    ''',
    roles = [
        {
            'key': 'personne_de_reference',
            'label': 'Personne de référence',
            'max': 1
            },
        {
            'key': 'conjoint',
            'label': 'Conjoint',
            'max': 1
            },
        {
            'key': 'enfant',
            'plural': 'enfants',
            'label': 'Enfants',
            },
        {
            'key': 'autre',
            'plural': 'autres',
            'label': 'Autres'
            }
        ]
    )

entities = [Individu, Famille, FoyerFiscal, Menage]



from typing import Callable, Tuple

import pytest

from ..cache import tax_benefit_system
from openfisca_core.tools import assert_near
from openfisca_core.simulations import Simulation
from openfisca_france.reforms.allocations_familiales_imposables import (
    allocations_familiales_imposables,
    )


@pytest.fixture
def simulations(new_scenario) -> Callable[..., Tuple[Simulation]]:
    def _simulations(year: int) -> Tuple[Simulation]:
        scenario = new_scenario(
            reform = allocations_familiales_imposables(tax_benefit_system),
            count = 10,
            _max = 30000,
            _min = 0,
            name = 'salaire_imposable',
            year = year,
            people = 4,
            )

        return scenario.new_simulation(use_baseline = True), scenario.new_simulation()

    return _simulations


def test_af(simulations, year = 2012, error = 0.01):
    actual_simulation, reform_simulation = simulations(year)
    actual_af = actual_simulation.calculate_add('af', year)
    reform_af = reform_simulation.calculate_add('af', year)

    assert_near(actual_af, reform_af, absolute_error_margin = error)


def test_af_imposables(simulations, year = 2012, error = 0.01):
    actual_simulation, reform_simulation = simulations(year)
    actual_af = actual_simulation.calculate_add('af', year)
    reform_af_imposables = reform_simulation.calculate(
        'allocations_familiales_imposables',
        year,
        )

    assert_near(actual_af, reform_af_imposables, absolute_error_margin = error)


def test_rbg(simulations, year = 2012, error = 0.01):
    actual_simulation, reform_simulation = simulations(year)
    actual_af = actual_simulation.calculate_add('af', year)
    actual_rbg = actual_simulation.calculate('rbg', year)
    reform_rbg = reform_simulation.calculate('rbg', year)

    assert_near(actual_af + actual_rbg, reform_rbg, absolute_error_margin = error)



from numpy import datetime64, timedelta64

from openfisca_france.model.base import *
from openfisca_france.model.caracteristiques_socio_demographiques.demographie import RegimeSecuriteSociale
from openfisca_france.model.prelevements_obligatoires.prelevements_sociaux.cotisations_sociales.base import apply_bareme_for_relevant_type_sal


class jei_date_demande(Variable):
    value_type = date
    default_value = date(2099, 12, 31)
    entity = Individu
    label = "Date de demande (et d'octroi) du statut de jeune entreprise innovante (JEI)"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class exoneration_cotisations_employeur_tode_eligibilite(Variable):
    value_type = bool
    entity = Individu
    label = "Éligibilité à l'exonération de cotisations employeur agricole pour travailleur occasionnel demandeur d'emploi (TO-DE)"
    reference = [
        'https://www.legifrance.gouv.fr/codes/article_lc/LEGIARTI000037947610/',
        'https://www.msa.fr/lfp/employeur/exonerations-travailleurs-occasionnels'
        ]
    definition_period = MONTH
    set_input = set_input_dispatch_by_period
    end = '2025-12-31'
    documentation = '''
        Non modélisé (2022), tout employeur MSA sauf ces employeurs :
        Coopératives d'utilisation de matériel agricole (CUMA).
        Coopératives de transformation, conditionnement et commercialisation.
        Entreprises paysagistes.
        Structures exerçant des activités de tourisme à la ferme.
        Entreprises de service (Crédit agricole, Groupama, caisses de MSA, groupements professionnels agricoles, Chambres d'agriculture…).
        Artisans ruraux.
        Entreprises de travail temporaire (ETT) et les entreprises de travail temporaire d'insertion (ETTI).
        Entreprises de travaux agricoles, ruraux et forestiers (ETARF).
    '''

    def formula_2019(individu, period):
        # employeur relevant de la MSA
        secteur_agricole = individu('secteur_activite_employeur', period) == TypesSecteurActivite.agricole
        regime_agricole = individu('regime_securite_sociale', period) == RegimeSecuriteSociale.regime_agricole

        # salarié travailleur occasionnel agricole
        travailleur_occasionnel_agricole = individu('travailleur_occasionnel_agricole', period)

        return (secteur_agricole + regime_agricole) * travailleur_occasionnel_agricole


class exoneration_cotisations_employeur_tode(Variable):
    value_type = float
    entity = Individu
    label = "Exonération de cotisations employeur agricole pour travailleur occasionnel demandeur d'emploi (TO-DE)"
    reference = [
        'https://www.legifrance.gouv.fr/codes/article_lc/LEGIARTI000037947610/',
        'https://www.legifrance.gouv.fr/codes/article_lc/LEGIARTI000038026966'
        ]
    definition_period = MONTH
    set_input = set_input_divide_by_period
    end = '2025-12-31'
    documentation = '''
        Exonération de cotisations et contributions employeur sur les bas salaires.

        Non modélisé (2022):
        La durée maximale d’application de l’exonération TO-DE est fixée à 119 jours
        consécutifs ou non, par employeur, par salarié et par année civile.
    '''

    def formula_2019(individu, period, parameters):
        # l'individu est le travailleur occasionnel
        eligible = individu('exoneration_cotisations_employeur_tode_eligibilite', period)

        # cotisations assurances sociales agricoles (ASA) - identiques régime général
        mmid_employeur = individu('mmid_employeur', period)
        allegement_cotisation_maladie_base = individu('allegement_cotisation_maladie_base', period)  # si rémunération <= 2.5 smic
        cotisations_asa = mmid_employeur + allegement_cotisation_maladie_base

        famille = individu('famille', period)
        accident_du_travail = individu('accident_du_travail', period)
        fnal = individu('fnal', period)

        vieillesse_deplafonnee_employeur = individu('vieillesse_deplafonnee_employeur', period)
        vieillesse_plafonnee_employeur = individu('vieillesse_plafonnee_employeur', period)
        agirc_arrco_employeur = individu('agirc_arrco_employeur', period)
        contribution_equilibre_general_employeur = individu('contribution_equilibre_general_employeur', period)

        contribution_solidarite_autonomie = individu('contribution_solidarite_autonomie', period)
        chomage_employeur = individu('chomage_employeur', period)

        # les cotisations sont des prélèvements, l'exonération leur opposé
        assiette_exoneration = -1 * (
            cotisations_asa
            + famille
            + accident_du_travail
            + fnal
            + vieillesse_deplafonnee_employeur
            + vieillesse_plafonnee_employeur
            + agirc_arrco_employeur
            + contribution_equilibre_general_employeur
            + contribution_solidarite_autonomie
            + chomage_employeur
            )

        # Exonération totale à <= 1.2 SMIC
        # Puis dégressive : 1,2 × C/0,40 × (1,6 × montant mensuel du SMIC/ rémunération mensuelle brute hors heures supplémentaires et complémentaires-1)
        # Devient nulle à >= 1.6 SMIC

        salaire_de_base = individu('salaire_de_base', period)
        smic_proratise = individu('smic_proratise', period)

        parameters_tode = parameters(period).prelevements_sociaux.reductions_cotisations_sociales.agricole.tode
        coefficient_degressivite = parameters_tode.plafond - parameters_tode.plafond_exoneration_integrale
        exoneration_degressive = parameters_tode.plafond_exoneration_integrale * (assiette_exoneration / coefficient_degressivite) * (parameters_tode.plafond * smic_proratise / salaire_de_base - 1)

        sous_plancher = salaire_de_base <= (parameters_tode.plafond_exoneration_integrale * smic_proratise)
        sous_plafond = salaire_de_base < (parameters_tode.plafond * smic_proratise)
        exoneration = where(sous_plancher, assiette_exoneration, sous_plafond * exoneration_degressive)

        # non cumul avec l'allègement général de cotisations employeur sur les bas salaires (Fillon)
        choix_exoneration_cotisations_employeur_agricole = individu('choix_exoneration_cotisations_employeur_agricole', period)

        return choix_exoneration_cotisations_employeur_agricole * eligible * exoneration


class choix_exoneration_cotisations_employeur_agricole(Variable):
    value_type = bool
    default_value = False
    entity = Individu
    label = "L'employeur agricole choisit une exonération de cotisations employeur spécifique au secteur agricole"
    reference = 'https://www.msa.fr/lfp/employeur/exonerations-travailleurs-occasionnels'
    definition_period = MONTH
    set_input = set_input_dispatch_by_period
    documentation = '''
    Pour un travailleur occasionnel, l'employeur agricole a le choix
    entre la réduction générale de cotisations sur les bas salaires (Fillon)
    et la TO-DE.
    La TO-DE est plus avantageuse mais à son arrêt au 12.2022,
    la réduction Fillon sera applicable.
    '''


class exoneration_cotisations_employeur_geographiques(Variable):
    value_type = float
    entity = Individu
    label = "Exonérations de cotisations employeur dépendant d'une zone géographique"
    reference = 'https://www.apce.com/pid815/aides-au-recrutement.html?espace=1&tp=1'
    definition_period = MONTH
    set_input = set_input_divide_by_period

    def formula(individu, period, parameters):
        exoneration_cotisations_employeur_zfu = individu('exoneration_cotisations_employeur_zfu', period, options = [ADD])
        exoneration_cotisations_employeur_zrd = individu('exoneration_cotisations_employeur_zrd', period, options = [ADD])
        exoneration_cotisations_employeur_zrr = individu('exoneration_cotisations_employeur_zrr', period, options = [ADD])

        exonerations_geographiques = (
            exoneration_cotisations_employeur_zfu
            + exoneration_cotisations_employeur_zrd
            + exoneration_cotisations_employeur_zrr
            )

        return exonerations_geographiques


class exoneration_cotisations_employeur_jei(Variable):
    value_type = float
    entity = Individu
    label = 'Exonération de cotisations employeur pour JEI (jeune entreprise innovante)'
    reference = 'http://www.apce.com/pid1653/jeune-entreprise-innovante.html?pid=1653&pagination=2'
    definition_period = MONTH
    set_input = set_input_divide_by_period

    def formula(individu, period, parameters):
        assiette_allegement = individu('assiette_allegement', period)
        jei_date_demande = individu('jei_date_demande', period)
        jeune_entreprise_innovante = individu('jeune_entreprise_innovante', period)
        plafond_securite_sociale = individu('plafond_securite_sociale', period)
        smic_proratise = individu('smic_proratise', period)
        categorie_salarie = individu('categorie_salarie', period)

        bareme_by_type_sal_name = parameters(period).cotsoc.cotisations_employeur
        bareme_names = ['vieillesse_deplafonnee', 'vieillesse_plafonnee', 'maladie', 'famille']

        exoneration = smic_proratise * 0.0
        for bareme_name in bareme_names:
            exoneration += apply_bareme_for_relevant_type_sal(
                bareme_by_type_sal_name = bareme_by_type_sal_name,
                bareme_name = bareme_name,
                categorie_salarie = categorie_salarie,
                base = min_(assiette_allegement, 4.5 * smic_proratise),
                plafond_securite_sociale = plafond_securite_sociale,
                round_base_decimals = 2,
                )

        exoneration_relative_year_passed = exoneration_relative_year(period, jei_date_demande)
        rate_by_year_passed = {
            0: 1,
            1: 1,
            2: 1,
            3: 1,
            4: 1,
            5: 1,
            6: 1,
            7: 1,
            }  # TODO: move to parameters file
        for year_passed, rate in rate_by_year_passed.items():
            condition_on_year_passed = exoneration_relative_year_passed == timedelta64(year_passed, 'Y')
            if condition_on_year_passed.any():
                exoneration[condition_on_year_passed] = rate * exoneration

        return - exoneration * jeune_entreprise_innovante


class exoneration_cotisations_employeur_zfu(Variable):
    value_type = float
    entity = Individu
    label = "Exonération de cotisations employeur pour l'embauche en ZFU (zone franche urbaine)"
    reference = 'http://www.apce.com/pid553/exoneration-dans-les-zfu.html?espace=1&tp=1&pagination=2'
    definition_period = MONTH
    set_input = set_input_divide_by_period

# TODO
# Ce dispositif d'exonération sociale est fermé depuis le 1er janvier 2015 mais reste applicable aux entreprises qui
# en bénéficiaient avant cette date.
# - ne pas être détenues à plus de 25 % par des entreprises employant plus de 250 salariés et dont le chiffre d'affaires
#   ou dont le bilan excède 50 M€ ou 43 M€,
# - disposer des éléments d'exploitation ou des stocks nécessaires à l'activité des salariés,
# - être à jour de ses cotisations sociales ou avoir souscrit à un engagement d'apurement progressif de ses dettes.
#
# Secteurs d'activité concernés
#
# L'exonération est applicable, quel que soit le secteur d'activité.
# Toutefois, les entreprises exerçant une activité principale dans les secteurs de la construction automobile,
# construction navale, fabrication de fibres textiles artificielles ou synthétiques, sidérurgie ou des transports
# routiers de marchandises, ne pourront pas bénéficier de cette exonération.

# Embauche de résidents (clause d'embauche locale)
# Pour les entreprises qui se créent ou s'implantent dans une ZFU à compter du 1er janvier 2012,  le bénéfice de
# l'exonération des cotisations sociales est subordonnée lors de toute nouvelle embauche à la condition que la moitié
# de salariés embauchés ou employés résident en ZFU ou en zone urbaine sensible.
#
# Le respect de la condition d'embauche locale est apprécié à la date d'effet de la nouvelle embauche dès la deuxième
# embauche.
#
# Précision : les salariés employés sont ceux déjà présents dans l'entreprise à la date de la nouvelle embauche, les
# salariés embauchés sont ceux recrutés depuis la date de création ou d'implantation de l'entreprise en ZFU.
#
# Est considéré comme résident le salarié habitant soit dans la ZFU d'implantation, soit dans l'une des ZUS de l'unité
# urbaine où se trouve la ZFU. Le maire peut, à la demande de l'employeur, fournir des éléments d'informations relatifs
# à la qualité de résident dans la zone afin de déterminer si la proportion exigée est respectée.
#
# Si la proportion n'est pas respectée à la date d'effet de l'embauche, l'employeur dispose d'un délai de 3 mois pour
# régulariser la situation. A défaut, le bénéfice de l'exonération est suspendu du 1er jour du mois suivant
# l'expiration du délai de 3 mois, jusqu'au 1er jour du mois suivant la date où la condition est de nouveau remplie.
#
# Le salarié résident doit être titulaire d'un contrat à durée indéterminée ou d'un contrat à durée déterminée d'au
# moins 12 mois, conclu pour une durée minimale de 16 heures par semaine.
# 5 ans +
# Dans les entreprises de 5 salariés et plus, les cotisations employeur bénéficient d'un abattement sur la base
# imposable pendant 3 ans de :
# - 60 % la première année,
# - 40 % la seconde année,
# - 20 % la troisième année.
#
# Dans les entreprises de moins de 5 salariés, un abattement est appliqué sur 9 ans de la manière suivante :
# - 60 % les 5 premières années,
# - 40 % les 2 années suivantes,
# - 20 % les deux dernières années.
#
# Le cumul de l'ensemble des aides publiques de minimis (allégements fiscaux, sociaux et aides des collectivités
# territoriales) ne peut dépasser le plafond des aides de minimis, fixé à 200 000 euros sur une période glissante de 36
# mois (100 000 euros pour les entreprises de transport routier).

    def formula(individu, period, parameters):
        assiette_allegement = individu('assiette_allegement', period)
        contrat_de_travail_type = individu('contrat_de_travail_type', period)
        TypesContrat = contrat_de_travail_type.possible_values
        contrat_de_travail_debut = individu('contrat_de_travail_debut', period)
        contrat_de_travail_fin = individu('contrat_de_travail_fin', period)
        effectif_entreprise = individu('effectif_entreprise', period)
        entreprise_chiffre_affaire = individu('entreprise_chiffre_affaire', period)
        entreprise_bilan = individu('entreprise_bilan', period)
        taux_versement_transport = individu('taux_versement_transport', period)

        # TODO: move to parameters file
        entreprise_eligible = (entreprise_chiffre_affaire <= 1e7) | (entreprise_bilan <= 1e7)

        smic_proratise = individu('smic_proratise', period)
        zone_franche_urbaine = individu('zone_franche_urbaine', period)

        duree_cdd_eligible = (contrat_de_travail_fin > contrat_de_travail_debut + timedelta64(365, 'D'))
        # TODO: move to parameters file
        contrat_de_travail_eligible = (contrat_de_travail_debut <= datetime64('2014-12-31')) * ((contrat_de_travail_type == TypesContrat.cdi) + ((contrat_de_travail_type == TypesContrat.cdd) * (duree_cdd_eligible)))
        # TODO: move to parameters file

        eligible = (
            contrat_de_travail_eligible
            * (effectif_entreprise <= 50)
            * zone_franche_urbaine
            * entreprise_eligible
            )

        bareme_by_name = parameters(period).cotsoc.cotisations_employeur['prive_non_cadre']

        if period.start.year < 2007:
            fnal_contrib = parameters(period).prelevements_sociaux.autres_taxes_participations_assises_salaires.fnal.contribution_plus_de_10_salaries
            fnal_contrib_seuil = 10
        elif period.start.year >= 2007 and period.start.year < 2020:
            fnal_contrib = parameters(period).prelevements_sociaux.autres_taxes_participations_assises_salaires.fnal.contribution_plus_de_20_salaries
            fnal_contrib_seuil = 20
        else:
            fnal_contrib = parameters(period).prelevements_sociaux.autres_taxes_participations_assises_salaires.fnal.contribution_plus_de_50_salaries
            fnal_contrib_seuil = 50

        if period.start.year < 2019:
            taux_maladie = bareme_by_name['maladie'].rates[0]
        else:
            taux_maladie = 0

        taux_max = (
            bareme_by_name['vieillesse_deplafonnee'].rates[0]
            + bareme_by_name['vieillesse_plafonnee'].rates[0]
            + taux_maladie
            + bareme_by_name['famille'].rates[0]
            + parameters(period).prelevements_sociaux.autres_taxes_participations_assises_salaires.fnal.cotisation.rates[0]
            + fnal_contrib.rates[0] * (effectif_entreprise >= fnal_contrib_seuil)
            + taux_versement_transport
            )

        # TODO: move to parameters file : voir http://www.urssaf.fr/images/ref_lc2009-077.pdf
        seuil_max = 2
        seuil_min = 1.4

        taux_exoneration = compute_taux_exoneration(assiette_allegement, smic_proratise, taux_max, seuil_max, seuil_min)
        exoneration_relative_year_passed = exoneration_relative_year(period, contrat_de_travail_debut)

        large_rate_by_year_passed = {
            0: 1,
            1: 1,
            2: 1,
            3: 1,
            4: 1,
            5: .60,
            6: .40,
            7: .20,
            }  # TODO: move to parameters file

        small_rate_by_year_passed = {
            0: 1,
            1: 1,
            2: 1,
            3: 1,
            4: 1,
            5: .60,
            6: .60,
            7: .60,
            8: .60,
            9: .60,
            10: .40,
            11: .40,
            12: .20,
            13: .20,
            }  # TODO: move to parameters file

        large_taux_exoneration = eligible * 0.0
        small_taux_exoneration = eligible * 0.0

        for year_passed, rate in large_rate_by_year_passed.items():
            condition_on_year_passed = exoneration_relative_year_passed == timedelta64(year_passed, 'Y')
            if condition_on_year_passed.any():
                large_taux_exoneration[condition_on_year_passed] = rate * taux_exoneration

        for year_passed, rate in small_rate_by_year_passed.items():
            condition_on_year_passed = exoneration_relative_year_passed == timedelta64(year_passed, 'Y')
            if condition_on_year_passed.any():
                small_taux_exoneration[condition_on_year_passed] = rate * taux_exoneration

        exoneration_cotisations_zfu = (
            eligible
            * assiette_allegement
            * (
                small_taux_exoneration
                * (effectif_entreprise <= 5)
                + large_taux_exoneration
                * (effectif_entreprise > 5)
                )
            )

        return exoneration_cotisations_zfu
        # TODO: propager dans le temps


class exoneration_cotisations_employeur_zrd(Variable):
    value_type = float
    entity = Individu
    label = "Exonération de cotisations employeur pour l'embauche en ZRD (zone de restructuration de la Défense)"
    reference = 'http://www.apce.com/pid11668/exoneration-dans-les-zrd.html?espace=1&tp=1'
    definition_period = MONTH
    set_input = set_input_divide_by_period

    def formula(individu, period, parameters):
        assiette_allegement = individu('assiette_allegement', period)
        entreprise_creation = individu('entreprise_creation', period)
        smic_proratise = individu('smic_proratise', period)
        zone_restructuration_defense = individu('zone_restructuration_defense', period)

        eligible = zone_restructuration_defense
        taux_max = .281  # TODO: move to parameters file
        seuil_max = 2.4
        seuil_min = 1.4
        taux_exoneration = compute_taux_exoneration(assiette_allegement, smic_proratise, taux_max, seuil_max, seuil_min)

        exoneration_relative_year_passed = exoneration_relative_year(period, entreprise_creation)
        rate_by_year_passed = {
            0: 1,
            1: 1,
            2: 1,
            3: 2 / 3,
            4: 1 / 3,
            }  # TODO: move to parameters file
        ratio = eligible * 0.0
        for year_passed, rate in rate_by_year_passed.items():
            condition_on_year_passed = exoneration_relative_year_passed == timedelta64(year_passed, 'Y')
            if condition_on_year_passed.any():
                ratio[condition_on_year_passed] = rate

        exoneration_cotisations_zrd = ratio * taux_exoneration * assiette_allegement * eligible

        return exoneration_cotisations_zrd


class exoneration_cotisations_employeur_zrr(Variable):
    value_type = float
    entity = Individu
    label = "Exonération de cotisations employeur pour l'embauche en ZRR (zone de revitalisation rurale)"
    reference = 'http://www.apce.com/pid538/embauches-en-zru-et-zrr.html?espace=1&tp=1'
    definition_period = MONTH
    set_input = set_input_divide_by_period

    # Les entreprises et groupements d'employeurs exerçant une activité industrielle, commerciale, artisanale, agricole
    # ou libérale et cotisant au régime d'assurance chômage.
    # Les entreprises concernées, y compris chacune de celles appartenant à un groupement d'employeurs, doivent avoir
    # au moins un établissement situé en zone de revitalisation rurale.
    #
    # A noter : les associations à but non lucratif sont exclues du dispositif. Par contre, quelle que soit leur forme
    # juridique, les entreprises d'insertion ou d'intérim d'insertion peuvent en bénéficier. Les régies de quartier
    # peuvent en bénéficier lorsque leur activité est susceptible d'entraîner l'assujettissement à la TVA à l'impôt sur
    # les sociétés ainsi qu'à la contribution économique territoriale qu'elles en soient effectivement redevables
    # ou non.
    #
    # L'employeur ne doit avoir procédé à aucun licenciement économique durant les 12 mois précédant l'embauche.

    def formula(individu, period, parameters):
        assiette_allegement = individu('assiette_allegement', period)
        contrat_de_travail_type = individu('contrat_de_travail_type', period)
        TypesContrat = contrat_de_travail_type.possible_values
        contrat_de_travail_debut = individu('contrat_de_travail_debut', period)
        contrat_de_travail_fin = individu('contrat_de_travail_fin', period)
        effectif_entreprise = individu('effectif_entreprise', period)
        smic_proratise = individu('smic_proratise', period)
        zone_revitalisation_rurale = individu('zone_revitalisation_rurale', period)

        duree_cdd_eligible = contrat_de_travail_fin > contrat_de_travail_debut + timedelta64(365, 'D')
        # TODO: move to parameters file
        contrat_de_travail_eligible = (contrat_de_travail_type == TypesContrat.cdi) + ((contrat_de_travail_type == TypesContrat.cdd) * (duree_cdd_eligible))

        duree_validite = (
            datetime64(period.start) + timedelta64(1, 'D') - contrat_de_travail_debut
            ).astype('timedelta64[Y]') < timedelta64(1, 'Y')

        eligible = (
            contrat_de_travail_eligible
            * (effectif_entreprise <= 50)
            * zone_revitalisation_rurale
            * duree_validite
            )

        taux_max = .281 if period.start.year < 2015 else .2655  # TODO: move to parameters file
        seuil_max = 2.4
        seuil_min = 1.5
        taux_exoneration = compute_taux_exoneration(assiette_allegement, smic_proratise, taux_max, seuil_max, seuil_min)
        exoneration_cotisations_zrr = taux_exoneration * assiette_allegement * eligible

        return exoneration_cotisations_zrr


# Aides à la création
class exoneration_is_creation_zrr(Variable):
    value_type = float
    entity = Individu
    label = "Exonérations fiscales pour création d'une entreprise en zone de revitalisation rurale (ZRR)"
    reference = 'http://www.apce.com/pid11690/exonerations-d-impots-zrr.html?espace=1&tp=1'
    definition_period = YEAR
    calculate_output = calculate_output_divide

    def formula(individu, period, parameters):
        decembre = period.first_month.offset(11, 'month')
        effectif_entreprise = individu('effectif_entreprise', decembre)
        entreprise_benefice = individu('entreprise_benefice', period, options = [ADD])
        # TODO: MODIFIER avec création d'entreprise
        contrat_de_travail_type = individu('contrat_de_travail_type', decembre)

        TypesContrat = contrat_de_travail_type.possible_values

        contrat_de_travail_debut = individu('contrat_de_travail_debut', decembre)
        contrat_de_travail_fin = individu('contrat_de_travail_fin', decembre)
        duree_eligible = contrat_de_travail_fin > contrat_de_travail_debut + timedelta64(365, 'D')
        # TODO: move to parameters file
        contrat_de_travail_eligible = (contrat_de_travail_type == TypesContrat.cdi) + ((contrat_de_travail_type == TypesContrat.cdd) * (duree_eligible))
        zone_revitalisation_rurale = individu('zone_revitalisation_rurale', decembre)

        eligible = (
            contrat_de_travail_eligible
            * (effectif_entreprise <= 50)
            * zone_revitalisation_rurale
            )

        exoneration_relative_year_passed = exoneration_relative_year(period, contrat_de_travail_debut)

        rate_by_year_passed = {
            0: 1,
            1: 1,
            2: 1,
            3: 1,
            4: 1,
            5: .75,
            6: .50,
            7: .25,
            }  # TODO: move to parameters file
        taux_exoneraion = eligible * 0.0

        for year_passed, rate in rate_by_year_passed.items():
            condition_on_year_passed = exoneration_relative_year_passed == timedelta64(year_passed, 'Y')
            taux_exoneraion[condition_on_year_passed] = rate

        return taux_exoneraion * entreprise_benefice
        # TODO: mettre sur toutes les années


# # class bassin_emploi_redynamiser(Variable):
#     value_type = bool
#     entity = Individu
#     label = "L'entreprise est située danns un bassin d'emploi à redynamiser(BER)"
#     # La liste des bassins d'emploi à redynamiser a été fixée par le décret n°2007-228 du 20 février 2007.
#     # Actuellement, deux régions sont concernées : Champagne-Ardenne (zone d'emploi de la Vallée de la Meuse)
#     # et Midi-Pyrénées (zone d'emploi de Lavelanet).
#
#     def formula(individu, period, parameters):
#         effectif_entreprise = individu('effectif_entreprise', period)
#         return (effectif_entreprise >= 1) * False

class jeune_entreprise_innovante(Variable):
    value_type = bool
    entity = Individu
    label = "L'entreprise est une jeune entreprise innovante"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period

    def formula(individu, period, parameters):
        # Toute entreprise existante au 1er janvier 2004 ou créée entre le 1er janvier 2004 et le 31 décembre 2016 à
        # condition de remplir les conditions suivantes :
        #
        # avoir moins de 8 ans d'existence au moment de la demande
        #
        # être réellement nouvelle, c'est-à-dire ne pas avoir été créée dans le cadre d'une concentration,
        # d'une restructuration, d'une extension d'activité préexistante ou d'une reprise
        #
        # employer moins de 250 personnes au cours de l'exercice au titre duquel elle demande à bénéficier de ce statut
        #
        # réaliser un chiffre d'affaires inférieur à 50 M€  et disposer d'un total de bilan inférieur à 43 M€
        #
        # être indépendante, c'est-à-dire que son capital doit être détenu pour 50 % au minimum par :
        #
        # - des personnes physiques
        #
        # - une ou plusieurs autres JEI dont 50 % du capital au moins est détenu par des personnes physiques
        #
        # - des associations ou fondations reconnues d'utilité publique à caractère scientifique
        #
        # - des établissements de recherche et d'enseignement et leurs filiales
        #
        # - des structures d'investissement sous réserve qu'il n'y ait pas de lien de dépendance telles que des :
        #   -  fonds communs de placement dans l'innovation (FCPI)
        #   -  sociétés de capital-risque
        #   -  fonds d'investissement de proximité (FIP)
        #   -  sociétés de développement régional (SDR)
        #   -  sociétés financières d'innovation (SFI)
        #   -  sociétés unipersonnelles d'investissements à risques (SUIR).
        #
        # réaliser des dépenses de R§D représentant au moins 15 % des charges fiscalement déductibles au titre du même
        # exercice.
        effectif_entreprise = individu('effectif_entreprise', period)
        entreprise_bilan = individu('entreprise_bilan', period)
        entreprise_chiffre_affaire = individu('entreprise_chiffre_affaire', period)
        entreprise_creation = individu('entreprise_creation', period)
        # entreprise_depenses_rd =  individu('entreprise_depenses_rd', period)
        jei_date_demande = individu('jei_date_demande', period)
        # TODO: move to parameters file
        # entreprise_depenses_rd > .15 TODO
        independance = True

        jeune_entreprise_innovante = (
            independance
            * (effectif_entreprise < 250)
            * (entreprise_creation <= datetime64('2016-12-31'))
            * ((jei_date_demande + timedelta64(1, 'D') - entreprise_creation).astype('timedelta64[Y]') < timedelta64(8, 'Y'))
            * (entreprise_chiffre_affaire < 50e6)
            * (entreprise_bilan < 43e6)
            )

        return jeune_entreprise_innovante


class bassin_emploi_redynamiser(Variable):
    value_type = bool
    entity = Individu
    label = "L'entreprise est située dans un bassin d'emploi à redynamiser (BER)"
    # La liste des bassins d'emploi à redynamiser a été fixée par le décret n°2007-228 du 20 février 2007.
    # Actuellement, deux régions sont concernées : Champagne-Ardenne (zone d'emploi de la Vallée de la Meuse)
    # et Midi-Pyrénées (zone d'emploi de Lavelanet).
    definition_period = MONTH
    set_input = set_input_dispatch_by_period

    def formula(individu, period, parameters):
        effectif_entreprise = individu('effectif_entreprise', period)

        return (effectif_entreprise >= 1) * False


class zone_restructuration_defense(Variable):
    value_type = bool
    entity = Individu
    label = "L'entreprise est située dans une zone de restructuration de la Défense (ZRD)"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period

    def formula(individu, period, parameters):
        effectif_entreprise = individu('effectif_entreprise', period)
        return (effectif_entreprise >= 1) * False


class zone_franche_urbaine(Variable):
    value_type = bool
    entity = Individu
    label = "L'entreprise est située danns une zone franche urbaine (ZFU)"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period

    def formula(individu, period, parameters):
        effectif_entreprise = individu('effectif_entreprise', period)
        return (effectif_entreprise >= 1) * False


class zone_revitalisation_rurale(Variable):
    value_type = bool
    entity = Individu
    label = "L'entreprise est située dans une zone de revitalisation rurale (ZRR)"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period

    def formula(individu, period, parameters):
        effectif_entreprise = individu('effectif_entreprise', period)
        return (effectif_entreprise >= 1) * False


# Helpers

def compute_taux_exoneration(assiette_allegement, smic_proratise, taux_max, seuil_max, seuil_min = 1):
    ratio_smic_salaire = smic_proratise / (assiette_allegement + 1e-16)
    # règle d'arrondi: 4 décimales au dix-millième le plus proche ( # TODO: reprise de l'allègement Fillon unchecked)
    return round_(
        taux_max * min_(1, max_(seuil_max * seuil_min * ratio_smic_salaire - seuil_min, 0) / (seuil_max - seuil_min)),
        4,
        )


def exoneration_relative_year(period, other_date):
    return (datetime64(period.start) + timedelta64(1, 'D') - other_date).astype('timedelta64[Y]')



from openfisca_core.periods import Period

from openfisca_france.model.base import *

from numpy import datetime64, absolute as abs_

# Références juridiques - Code de la sécurité sociale
#
# Article L821-1 / 821-8
# https://www.legifrance.gouv.fr/affichCode.do;jsessionid=0E604431776A4B1ED8D2F8EB55A1A99C.tplgfr35s_1?idSectionTA=LEGISCTA000006141693&cidTexte=LEGITEXT000006073189&dateTexte=20180412
#
# Article R821-1 / 821-9
# https://www.legifrance.gouv.fr/affichCode.do;jsessionid=0E604431776A4B1ED8D2F8EB55A1A99C.tplgfr35s_1?idSectionTA=LEGISCTA000006142017&cidTexte=LEGITEXT000006073189&dateTexte=20181010
#
# Article D821-1 / 821-11
# https://www.legifrance.gouv.fr/affichCode.do;jsessionid=157287C570B3AE9450A0BD88AA902970.tplgfr38s_1?idSectionTA=LEGISCTA000006141593&cidTexte=LEGITEXT000006073189&dateTexte=20180731


class aah_date_debut_incarceration(Variable):
    value_type = date
    default_value = date.max
    label = "La date de début d'incarcération"
    entity = Individu
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class aah_date_debut_hospitalisation(Variable):
    value_type = date
    default_value = date.max
    label = "La date de début d'hospitalisation"
    entity = Individu
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class aah_base_ressources(Variable):
    value_type = float
    label = "Base ressources de l'allocation adulte handicapé"
    entity = Individu
    definition_period = MONTH
    set_input = set_input_divide_by_period

    def formula(individu, period, parameters):  # formule pour les cas les plus anciens, moins vérifiée (sans abattement à part celui de 20% sur les revenus du conjoint)
        _parameters = parameters(period)
        aah = _parameters.prestations_sociales.prestations_etat_de_sante.invalidite.aah

        def assiette_conjoint(revenus_conjoint):
            return (1 - _parameters.impot_revenu.calcul_revenus_imposables.deductions.taux_salaires_pensions) * (1 - aah.abattement_conjoint.abattement_proportionnel) * revenus_conjoint

        def base_ressource_eval_annuelle():
            base_ressource = individu('aah_base_ressources_activite_eval_annuelle', period) + individu('aah_base_ressources_hors_activite_eval_annuelle', period)

            base_ressource_demandeur_conjoint = individu.famille.demandeur('aah_base_ressources_activite_eval_annuelle', period) + individu.famille.demandeur('aah_base_ressources_hors_activite_eval_annuelle', period)
            base_ressource_conjoint_conjoint = individu.famille.conjoint('aah_base_ressources_activite_eval_annuelle', period) + individu.famille.conjoint('aah_base_ressources_hors_activite_eval_annuelle', period)
            base_ressource_conjoint = base_ressource_conjoint_conjoint * individu.has_role(Famille.DEMANDEUR) + base_ressource_demandeur_conjoint * individu.has_role(Famille.CONJOINT)

            return base_ressource + assiette_conjoint(base_ressource_conjoint)

        return base_ressource_eval_annuelle() / 12

    def formula_2005_07_01(individu, period, parameters):
        law = parameters(period)
        aah = law.prestations_sociales.prestations_etat_de_sante.invalidite.aah

        def assiette_conjoint(revenus_conjoint):
            return (1 - law.impot_revenu.calcul_revenus_imposables.deductions.taux_salaires_pensions) * (1 - aah.abattement_conjoint.abattement_proportionnel) * revenus_conjoint

        def assiette_revenu_activite_demandeur(revenus_demandeur):
            smic_brut_horaire = law.marche_travail.salaire_minimum.smic.smic_b_horaire
            seuil1 = aah.travail_ordinaire.tranche_smic_horaire1 * smic_brut_horaire
            seuil2 = aah.travail_ordinaire.tranche_smic_horaire2 * smic_brut_horaire
            seuil3 = aah.travail_ordinaire.tranche_smic_horaire3 * smic_brut_horaire
            seuil4 = aah.travail_ordinaire.tranche_smic_horaire4 * smic_brut_horaire
            total_tranche1 = min_(seuil1, revenus_demandeur) * (1 - aah.travail_ordinaire.abattement_300)
            total_tranche2 = max_(0, min_(revenus_demandeur - seuil1, seuil2 - seuil1)) * (1 - aah.travail_ordinaire.abattement_700)
            total_tranche3 = max_(0, min_(revenus_demandeur - seuil2, seuil3 - seuil2)) * (1 - aah.travail_ordinaire.abattement_1100)
            total_tranche4 = max_(0, min_(revenus_demandeur - seuil3, seuil4 - seuil3)) * (1 - aah.travail_ordinaire.abattement_1500)
            total_tranche5 = max_(0, revenus_demandeur - seuil4)
            return total_tranche1 + total_tranche2 + total_tranche3 + total_tranche4 + total_tranche5

        def base_ressource_eval_annuelle():
            base_ressource_activite = assiette_revenu_activite_demandeur(individu('aah_base_ressources_activite_eval_annuelle', period))
            base_ressource = base_ressource_activite + individu('aah_base_ressources_hors_activite_eval_annuelle', period)

            base_ressource_demandeur_conjoint = individu.famille.demandeur('aah_base_ressources_activite_eval_annuelle', period) + individu.famille.demandeur('aah_base_ressources_hors_activite_eval_annuelle', period)
            base_ressource_conjoint_conjoint = individu.famille.conjoint('aah_base_ressources_activite_eval_annuelle', period) + individu.famille.conjoint('aah_base_ressources_hors_activite_eval_annuelle', period)
            base_ressource_conjoint = base_ressource_conjoint_conjoint * individu.has_role(Famille.DEMANDEUR) + base_ressource_demandeur_conjoint * individu.has_role(Famille.CONJOINT)

            return base_ressource + assiette_conjoint(base_ressource_conjoint)

        return base_ressource_eval_annuelle() / 12

    def formula_2011(individu, period, parameters):
        law = parameters(period)
        aah = law.prestations_sociales.prestations_etat_de_sante.invalidite.aah

        en_activite = (individu('salaire_imposable', period, options = [ADD]) + individu('rpns_imposables', period.last_year) > 0)

        def assiette_conjoint(revenus_conjoint):
            return (1 - law.impot_revenu.calcul_revenus_imposables.deductions.taux_salaires_pensions) * (1 - aah.abattement_conjoint.abattement_proportionnel) * revenus_conjoint

        def assiette_revenu_activite_demandeur(revenus_demandeur):
            smic_brut_annuel = 12 * law.marche_travail.salaire_minimum.smic.smic_b_horaire * law.marche_travail.salaire_minimum.smic.nb_heures_travail_mensuel
            total_tranche1 = min_(aah.travail_ordinaire.tranche_smic * smic_brut_annuel, revenus_demandeur)
            total_tranche2 = max_(0, revenus_demandeur - total_tranche1)
            return (1 - aah.travail_ordinaire.abattement_30) * total_tranche1 + (1 - aah.travail_ordinaire.abattement_sup) * total_tranche2

        def base_ressource_eval_trim():
            three_previous_months = Period(('month', period.first_month.start, 3)).offset(-3)
            base_ressource_activite = individu('aah_base_ressources_activite_eval_trimestrielle', period) - individu('aah_base_ressources_activite_milieu_protege', three_previous_months, options = [ADD])
            base_ressource_hors_activite = individu('aah_base_ressources_hors_activite_eval_trimestrielle', period) + individu('aah_base_ressources_activite_milieu_protege', three_previous_months, options = [ADD])

            base_ressource_demandeur = max_(0, assiette_revenu_activite_demandeur(base_ressource_activite) + base_ressource_hors_activite)

            base_ressource_demandeur_conjoint = max_(0, individu.famille.demandeur('aah_base_ressources_activite_eval_trimestrielle', period) + individu.famille.demandeur('aah_base_ressources_hors_activite_eval_trimestrielle', period))
            base_ressource_conjoint_conjoint = max_(0, individu.famille.conjoint('aah_base_ressources_activite_eval_trimestrielle', period) + individu.famille.conjoint('aah_base_ressources_hors_activite_eval_trimestrielle', period))
            base_ressource_conjoint = base_ressource_conjoint_conjoint * individu.has_role(Famille.DEMANDEUR) + base_ressource_demandeur_conjoint * individu.has_role(Famille.CONJOINT)

            return base_ressource_demandeur + assiette_conjoint(base_ressource_conjoint)

        def base_ressource_eval_annuelle():
            base_ressource_activite = assiette_revenu_activite_demandeur(individu('aah_base_ressources_activite_eval_annuelle', period))
            base_ressource = base_ressource_activite + individu('aah_base_ressources_hors_activite_eval_annuelle', period)

            base_ressource_demandeur_conjoint = individu.famille.demandeur('aah_base_ressources_activite_eval_annuelle', period) + individu.famille.demandeur('aah_base_ressources_hors_activite_eval_annuelle', period)
            base_ressource_conjoint_conjoint = individu.famille.conjoint('aah_base_ressources_activite_eval_annuelle', period) + individu.famille.conjoint('aah_base_ressources_hors_activite_eval_annuelle', period)
            base_ressource_conjoint = base_ressource_conjoint_conjoint * individu.has_role(Famille.DEMANDEUR) + base_ressource_demandeur_conjoint * individu.has_role(Famille.CONJOINT)

            return base_ressource + assiette_conjoint(base_ressource_conjoint)

        return where(
            en_activite,
            base_ressource_eval_trim() / 12,
            base_ressource_eval_annuelle() / 12
            )

    def formula_2022_01_01(individu, period, parameters):
        law = parameters(period)
        aah = law.prestations_sociales.prestations_etat_de_sante.invalidite.aah

        en_activite = ((individu('salaire_imposable', period, options = [ADD]) + individu('rpns_imposables', period.last_year) > 0))

        def assiette_conjoint(revenus_conjoint):
            af_nbenf = individu.famille('af_nbenf', period)
            revenus = (1 - law.impot_revenu.calcul_revenus_imposables.deductions.taux_salaires_pensions) * revenus_conjoint
            return max_(revenus - (aah.abattement_conjoint.abattement_forfaitaire.base + aah.abattement_conjoint.abattement_forfaitaire.majoration_pac * af_nbenf), 0)

        def assiette_revenu_activite_demandeur(revenus_demandeur):
            smic_brut_annuel = 12 * law.marche_travail.salaire_minimum.smic.smic_b_horaire * law.marche_travail.salaire_minimum.smic.nb_heures_travail_mensuel
            total_tranche1 = min_(aah.travail_ordinaire.tranche_smic * smic_brut_annuel, revenus_demandeur)
            total_tranche2 = max_(0, revenus_demandeur - total_tranche1)
            return (1 - aah.travail_ordinaire.abattement_30) * total_tranche1 + (1 - aah.travail_ordinaire.abattement_sup) * total_tranche2

        def base_ressource_eval_trim():
            three_previous_months = Period(('month', period.first_month.start, 3)).offset(-3)
            base_ressource_activite = individu('aah_base_ressources_activite_eval_trimestrielle', period) - individu('aah_base_ressources_activite_milieu_protege', three_previous_months, options = [ADD])
            base_ressource_hors_activite = individu('aah_base_ressources_hors_activite_eval_trimestrielle', period) + individu('aah_base_ressources_activite_milieu_protege', three_previous_months, options = [ADD])

            base_ressource_demandeur = max_(0, assiette_revenu_activite_demandeur(base_ressource_activite) + base_ressource_hors_activite)

            base_ressource_demandeur_conjoint = individu.famille.demandeur('aah_base_ressources_activite_eval_trimestrielle', period) + individu.famille.demandeur('aah_base_ressources_hors_activite_eval_trimestrielle', period)
            base_ressource_conjoint_conjoint = max_(0, individu.famille.conjoint('aah_base_ressources_activite_eval_trimestrielle', period) + individu.famille.conjoint('aah_base_ressources_hors_activite_eval_trimestrielle', period))
            base_ressource_conjoint = base_ressource_conjoint_conjoint * individu.has_role(Famille.DEMANDEUR) + base_ressource_demandeur_conjoint * individu.has_role(Famille.CONJOINT)

            return base_ressource_demandeur + assiette_conjoint(base_ressource_conjoint)

        def base_ressource_eval_annuelle():
            base_ressource_activite = assiette_revenu_activite_demandeur(individu('aah_base_ressources_activite_eval_annuelle', period))
            base_ressource = base_ressource_activite + individu('aah_base_ressources_hors_activite_eval_annuelle', period)

            base_ressource_demandeur_conjoint = individu.famille.demandeur('aah_base_ressources_activite_eval_annuelle', period) + individu.famille.demandeur('aah_base_ressources_hors_activite_eval_annuelle', period)
            base_ressource_conjoint_conjoint = individu.famille.conjoint('aah_base_ressources_activite_eval_annuelle', period) + individu.famille.conjoint('aah_base_ressources_hors_activite_eval_annuelle', period)
            base_ressource_conjoint = base_ressource_conjoint_conjoint * individu.has_role(Famille.DEMANDEUR) + base_ressource_demandeur_conjoint * individu.has_role(Famille.CONJOINT)

            return base_ressource + assiette_conjoint(base_ressource_conjoint)

        return where(
            en_activite,
            base_ressource_eval_trim() / 12,
            base_ressource_eval_annuelle() / 12
            )

    def formula_2023_10_01(individu, period, parameters):
        law = parameters(period)
        aah = law.prestations_sociales.prestations_etat_de_sante.invalidite.aah

        en_activite = ((individu('salaire_imposable', period, options = [ADD]) + individu('rpns_imposables', period.last_year)) > 0)

        def assiette_revenu_activite_demandeur(revenus_demandeur):
            smic_brut_annuel = 12 * law.marche_travail.salaire_minimum.smic.smic_b_horaire * law.marche_travail.salaire_minimum.smic.nb_heures_travail_mensuel
            total_tranche1 = min_(aah.travail_ordinaire.tranche_smic * smic_brut_annuel, revenus_demandeur)
            total_tranche2 = max_(0, revenus_demandeur - total_tranche1)
            return (1 - aah.travail_ordinaire.abattement_30) * total_tranche1 + (1 - aah.travail_ordinaire.abattement_sup) * total_tranche2

        def base_ressource_eval_trim():
            three_previous_months = Period(('month', period.first_month.start, 3)).offset(-3)
            base_ressource_activite = individu('aah_base_ressources_activite_eval_trimestrielle', period) - individu('aah_base_ressources_activite_milieu_protege', three_previous_months, options = [ADD])
            base_ressource_hors_activite = individu('aah_base_ressources_hors_activite_eval_trimestrielle', period) + individu('aah_base_ressources_activite_milieu_protege', three_previous_months, options = [ADD])

            base_ressource_demandeur = max_(0, assiette_revenu_activite_demandeur(base_ressource_activite) + base_ressource_hors_activite)

            return base_ressource_demandeur

        def base_ressource_eval_annuelle():
            base_ressource_activite = assiette_revenu_activite_demandeur(individu('aah_base_ressources_activite_eval_annuelle', period))
            base_ressource = base_ressource_activite + individu('aah_base_ressources_hors_activite_eval_annuelle', period)

            return base_ressource

        return where(
            en_activite,
            base_ressource_eval_trim() / 12,
            base_ressource_eval_annuelle() / 12
            )
        # TODO: - Prendre en compte les abattements temporaires sur les ressources en cas de changement de situation (6 mois pour retour à l'emploi, un an pour inactivité)
        #       - La formule du calcul de la base de ressource est celle en vigueur à partir de 2011, avant 2011:
        #           - l'abattement pour les personnes invalides (défini dans l'art. 157 du CGI) sur le revenu net global est pris en compte (art. R821-4 du CSS)


class aah_base_ressources_activite_eval_trimestrielle(Variable):
    value_type = float
    label = "Base de ressources des revenus d'activité de l'AAH pour un individu, évaluation trimestrielle"
    entity = Individu
    definition_period = MONTH
    set_input = set_input_divide_by_period

    '''
        N'entrent pas en compte dans les ressources :
        L'allocation compensatrice tierce personne, les allocations familiales,
        l'allocation de logement, la retraite du combattant, les rentes viagères
        constituées en faveur d'une personne handicapée ou dans la limite d'un
        montant fixé à l'article D.821-6 du code de la sécurité sociale (1 830 €/an),
        lorsqu'elles ont été constituées par une personne handicapée pour elle-même.
        Le RMI (article R 531-10 du code de la sécurité sociale).
        A partir du 1er juillet 2007, votre Caf, pour le calcul de votre Aah,
        continue à prendre en compte les ressources de votre foyer diminuées de 20%.
        Notez, dans certaines situations, la Caf évalue forfaitairement vos
        ressources à partir de votre revenu mensuel.
    '''

    def formula(individu, period):
        period = period.first_month
        three_previous_months = Period(('month', period.start, 3)).offset(-3)
        last_year = period.last_year

        ressources_a_inclure = [
            'chomage_net',
            'indemnites_chomage_partiel',
            'indemnites_journalieres_imposables',
            'indemnites_stage',
            'revenus_stage_formation_pro',
            'salaire_net',
            ]

        ressources = sum(
            [individu(ressource, three_previous_months, options = [ADD]) for ressource in ressources_a_inclure]
            )

        def revenus_tns():
            revenus_auto_entrepreneur = individu('rpns_auto_entrepreneur_benefice', three_previous_months, options = [ADD])

            # Les revenus TNS hors AE sont estimés en se basant sur le revenu N-1
            rpns_micro_entreprise_benefice = individu('rpns_micro_entreprise_benefice', last_year) * 3 / 12
            rpns_benefice_exploitant_agricole = individu('rpns_benefice_exploitant_agricole', last_year) * 3 / 12
            rpns_autres_revenus = individu('rpns_autres_revenus', last_year) * 3 / 12

            return revenus_auto_entrepreneur + rpns_micro_entreprise_benefice + rpns_benefice_exploitant_agricole + rpns_autres_revenus

        return (ressources + revenus_tns()) * 4


class aah_base_ressources_activite_milieu_protege(Variable):
    value_type = float
    label = "Base de ressources de l'AAH des revenus d'activité en milieu protégé pour un individu"
    entity = Individu
    definition_period = MONTH
    set_input = set_input_divide_by_period


class aah_base_ressources_hors_activite_eval_trimestrielle(Variable):
    value_type = float
    label = "Base de ressources hors revenus d'activité de l'AAH pour un individu, évaluation trimestrielle"
    entity = Individu
    definition_period = MONTH
    set_input = set_input_divide_by_period

    '''
        N'entrent pas en compte dans les ressources :
        L'allocation compensatrice tierce personne, les allocations familiales,
        l'allocation de logement, la retraite du combattant, les rentes viagères
        constituées en faveur d'une personne handicapée ou dans la limite d'un
        montant fixé à l'article D.821-6 du code de la sécurité sociale (1 830 €/an),
        lorsqu'elles ont été constituées par une personne handicapée pour elle-même.
        Le RMI (article R 531-10 du code de la sécurité sociale).
        A partir du 1er juillet 2007, votre Caf, pour le calcul de votre Aah,
        continue à prendre en compte les ressources de votre foyer diminuées de 20%.
        Notez, dans certaines situations, la Caf évalue forfaitairement vos
        ressources à partir de votre revenu mensuel.
    '''

    def formula(individu, period):
        period = period.first_month
        three_previous_months = Period(('month', period.start, 3)).offset(-3)

        ressources_a_inclure = [
            'asi',
            'allocation_securisation_professionnelle',
            'bourse_recherche',
            'gains_exceptionnels',
            'pensions_alimentaires_percues',
            'pensions_invalidite',
            'prestation_compensatoire',
            'retraite_nette',
            'rsa_base_ressources_patrimoine_individu',
            ]

        ressources = sum(
            [individu(ressource, three_previous_months, options = [ADD]) for ressource in ressources_a_inclure]
            )
        # On récupère le montant absolu des pensions alimentaires versées au cas où la valeur reçue est négative
        pensions_alimentaires_versees = abs_(individu(
            'pensions_alimentaires_versees_individu', three_previous_months, options = [ADD]))

        # On soustrait le montant des pensions alimentaires versées à la base des ressources
        return (ressources - pensions_alimentaires_versees) * 4


class aah_base_ressources_activite_eval_annuelle(Variable):
    value_type = float
    label = "Base de ressources de l'AAH pour un individu, évaluation annuelle"
    entity = Individu
    definition_period = MONTH
    set_input = set_input_divide_by_period
    reference = [
        'Article R532-5 du Code de la sécurité sociale',
        'https://www.legifrance.gouv.fr/codes/article_lc/LEGIARTI000006750910'
        ]
    # TODO prendre en compte l'abattement sur les revenus de l'année précédente entre 2005 et 2010 lorsqu'une période d'inactivité
    # sans revenu de remplacement survient (https://www.legifrance.gouv.fr/codes/article_lc/LEGIARTI000006739692/2005-07-01/)
    # et les autres abattements sur reprise d'activité https://www.legifrance.gouv.fr/jorf/id/JORFTEXT000023086051/#JORFARTI000023086056 (notamment Art.D. 821-9 1°)

    def formula(individu, period, parameters):
        return (
            individu('salaire_imposable', period.n_2, options = [ADD])
            + individu('rpns_imposables', period.n_2)
            + individu('chomage_imposable', period.n_2, options = [ADD]))


class aah_base_ressources_hors_activite_eval_annuelle(Variable):
    value_type = float
    label = "Base de ressources de l'AAH pour un individu, évaluation annuelle"
    entity = Individu
    definition_period = MONTH
    set_input = set_input_divide_by_period
    reference = [
        'Article R532-5 du Code de la sécurité sociale',
        'https://www.legifrance.gouv.fr/codes/article_lc/LEGIARTI000006750910'
        ]
    # TODO prendre en compte l'abattement sur les revenus de l'année précédente entre 2005 et 2010 lorsqu'une période d'inactivité
    # sans revenu de remplacement survient (https://www.legifrance.gouv.fr/codes/article_lc/LEGIARTI000006739692/2005-07-01/)

    def formula(individu, period, parameters):
        return (individu('revenu_assimile_pension', period.n_2))


class aah_restriction_substantielle_durable_acces_emploi(Variable):
    value_type = bool
    default_value = True
    entity = Individu
    label = "Restriction substantielle et durable pour l'accès à l'emploi reconnue par la commission des droits et de l'autonomie des personnes handicapées"
    reference = [
        'Article L821-2 du Code de la sécurité sociale',
        'https://www.legifrance.gouv.fr/affichCodeArticle.do;jsessionid=17BE3036A19374AA1C8C7A4169702CD7.tplgfr24s_2?idArticle=LEGIARTI000020039305&cidTexte=LEGITEXT000006073189&dateTexte=20180731'
        ]
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class aah_eligible(Variable):
    value_type = bool
    label = "Eligibilité à l'Allocation adulte handicapé"
    entity = Individu
    definition_period = MONTH
    set_input = set_input_dispatch_by_period
    reference = ['Article 821-1 du Code de la sécurité sociale', 'https://www.legifrance.gouv.fr/codes/id/LEGIARTI000006739685/']

    '''
        Allocation adulte handicapé

        Conditions liées au handicap
        La personne doit être atteinte d’un taux d’incapacité permanente :
        - d’au moins 80 %,
        - ou compris entre 50 et 79 %. Dans ce cas, elle doit remplir deux conditions
        supplémentaires : être dans l’impossibilité de se procurer un emploi compte
        tenu de son handicap et ne pas avoir travaillé depuis au moins 1 an
        Condition de résidence
        L'AAH peut être versée aux personnes résidant en France métropolitaine ou
        dans les départements d'outre-mer ou à Saint-Pierre et Miquelon de façon permanente.
        Les personnes de nationalité étrangère doivent être en possession d'un titre de séjour
        régulier ou être titulaire d'un récépissé de renouvellement de titre de séjour.
        Condition d'âge
        Age minimum : Le demandeur ne doit plus avoir l'âge de bénéficier de l'allocation d'éducation de l'enfant
        handicapé, c'est-à-dire qu'il doit être âgé :
        - de plus de vingt ans,
        - ou de plus de seize ans, s'il ne remplit plus les conditions pour ouvrir droit aux allocations familiales.
        Pour les montants http://www.handipole.org/spip.php?article666

        Âge max_
        Le versement de l'AAH prend fin à partir de l'âge minimum légal de départ à la retraite en cas d'incapacité
        de 50 % à 79 %. À cet âge, le bénéficiaire bascule dans le régime de retraite pour inaptitude.
        En cas d'incapacité d'au moins 80 %, une AAH différentielle (c'est-à-dire une allocation mensuelle réduite)
        peut être versée au-delà de l'âge minimum légal de départ à la retraite en complément d'une retraite inférieure
        au minimum vieillesse.

        Avant 2011, l'allocation était perçue uniquement pour les personnes sans activité les douze mois précédents
        (ref https://www.legifrance.gouv.fr/codes/id/LEGIARTI000006739685/2005-06-30/),
        mais ce n'était pas le cas avant 2005 ref https://www.legifrance.gouv.fr/jorf/article_jo/JORFARTI000002101708
    '''

    def formula_2011_01_01(individu, period, parameters):
        law = parameters(period).prestations_sociales.prestations_etat_de_sante.invalidite.aah
        taux_incapacite = individu('taux_incapacite', period)
        taux_incapacite_max = (taux_incapacite >= law.taux_capacite.taux_incapacite)
        taux_incapacite_rsdae = (taux_incapacite >= law.taux_capacite.taux_incapacite_rsdae)
        rsdae = individu('aah_restriction_substantielle_durable_acces_emploi', period)

        age = individu('age', period)
        prestations_familiales_enfant_a_charge = individu('prestations_familiales_enfant_a_charge', period)
        eligible_aah = (
            (taux_incapacite_max + (taux_incapacite_rsdae * rsdae * (age <= law.age_legal_retraite)))
            * ((age >= law.age_minimal) + ((age >= law.age_fin_educ) * not_(prestations_familiales_enfant_a_charge)))
            )

        return eligible_aah

    def formula_2005_07_01(individu, period, parameters):
        law = parameters(period).prestations_sociales.prestations_etat_de_sante.invalidite.aah
        taux_incapacite = individu('taux_incapacite', period)
        taux_incapacite_max = (taux_incapacite >= law.taux_capacite.taux_incapacite)
        taux_incapacite_rsdae = (taux_incapacite >= law.taux_capacite.taux_incapacite_rsdae)
        rsdae = individu('aah_restriction_substantielle_durable_acces_emploi', period)

        age = individu('age', period)
        prestations_familiales_enfant_a_charge = individu('prestations_familiales_enfant_a_charge', period)
        eligible_aah = (
            (taux_incapacite_max + (taux_incapacite_rsdae * rsdae * (age <= law.age_legal_retraite))
            * (individu('salaire_imposable', period.last_year, options=[ADD]) <= 0))
            * ((age >= law.age_minimal) + ((age >= law.age_fin_educ) * not_(prestations_familiales_enfant_a_charge)))
            )
        return eligible_aah

    def formula(individu, period, parameters):
        law = parameters(period).prestations_sociales.prestations_etat_de_sante.invalidite.aah
        taux_incapacite = individu('taux_incapacite', period)
        taux_incapacite_max = (taux_incapacite >= law.taux_capacite.taux_incapacite)
        taux_incapacite_rsdae = (taux_incapacite >= law.taux_capacite.taux_incapacite_rsdae)
        rsdae = individu('aah_restriction_substantielle_durable_acces_emploi', period)

        age = individu('age', period)
        prestations_familiales_enfant_a_charge = individu('prestations_familiales_enfant_a_charge', period)
        eligible_aah = (
            (taux_incapacite_max + (taux_incapacite_rsdae * rsdae
            * (age <= law.age_legal_retraite)))
            * ((age >= law.age_minimal) + ((age >= law.age_fin_educ) * not_(prestations_familiales_enfant_a_charge)))
            )
        return eligible_aah


class aah_base_non_cumulable(Variable):
    value_type = float
    label = "Montant de l'Allocation adulte handicapé (hors complément) pour un individu, mensualisée"
    entity = Individu
    definition_period = MONTH
    set_input = set_input_divide_by_period

    def formula(individu, period):
        return individu('pensions_invalidite', period) + individu('asi', period.last_month)


class aah_plafond_ressources(Variable):
    value_type = float
    label = "Montant plafond des ressources pour bénéficier de l'Allocation adulte handicapé (hors complément)"
    entity = Individu
    reference = [
        'Article D821-2 du Code de la sécurité sociale',
        'https://www.legifrance.gouv.fr/affichCodeArticle.do;jsessionid=4B54EC7065520E4812F84677B918A48E.tplgfr28s_2?idArticle=LEGIARTI000019077584&cidTexte=LEGITEXT000006073189&dateTexte=20081218'
        ]
    definition_period = MONTH
    set_input = set_input_divide_by_period

    def formula(individu, period, parameters):
        law = parameters(period).prestations_sociales

        en_couple = individu.famille('en_couple', period)
        af_nbenf = individu.famille('af_nbenf', period)
        montant_max = law.prestations_etat_de_sante.invalidite.aah.montant

        return montant_max * (
            + 1
            + en_couple
            * law.prestations_etat_de_sante.invalidite.aah.majoration_plafond.majoration_plafond_couple
            + law.prestations_etat_de_sante.invalidite.aah.majoration_plafond.majoration_par_enfant_supplementaire
            * af_nbenf
            )


class aah_base(Variable):
    calculate_output = calculate_output_add
    value_type = float
    label = "Montant de l'Allocation adulte handicapé (hors complément) pour un individu, mensualisée"
    entity = Individu
    reference = [
        'Article L821-1 du Code de la sécurité sociale',
        'https://www.legifrance.gouv.fr/affichCodeArticle.do;jsessionid=53AFF5AA4010B01F0539052A33180B39.tplgfr35s_1?idArticle=LEGIARTI000033813790&cidTexte=LEGITEXT000006073189&dateTexte=20180412'
        " D'après service-public : 'Si vous touchez une pension ou une rente, vous recevez la différence entre le montant de votre pension ou rente et le montant maximal de l'AAH', le montant évoqué est donc le maximal"]
    definition_period = MONTH
    set_input = set_input_divide_by_period

    def formula(individu, period, parameters):
        law = parameters(period).prestations_sociales

        aah_eligible = individu('aah_eligible', period)
        aah_base_ressources = individu('aah_base_ressources', period)
        plaf_ress_aah = individu('aah_plafond_ressources', period)
        # Le montant de l'AAH est plafonné au montant de base.
        montant_max = law.prestations_etat_de_sante.invalidite.aah.montant
        montant_aah = min_(montant_max, max_(0, plaf_ress_aah - aah_base_ressources))

        aah_base_non_cumulable = individu('aah_base_non_cumulable', period)

        return aah_eligible * min_(max_(0, montant_aah), max_(0, montant_max - aah_base_non_cumulable))


class aah(Variable):
    calculate_output = calculate_output_add
    value_type = float
    label = 'Allocation adulte handicapé mensualisée'
    reference = 'https://www.legifrance.gouv.fr/affichCodeArticle.do?cidTexte=LEGITEXT000006073189&idArticle=LEGIARTI000006754198'
    entity = Individu
    definition_period = MONTH
    set_input = set_input_divide_by_period

    def formula(individu, period, parameters):
        aah_base = individu('aah_base', period)
        aah_parameters = parameters(period).prestations_sociales.prestations_etat_de_sante.invalidite.aah
        m_2 = datetime64(period.offset(-60, 'day').start)

        aah_date_debut_hospitalisation = individu('aah_date_debut_hospitalisation', period)
        aah_date_debut_incarceration = individu('aah_date_debut_incarceration', period)
        pers_charge = (individu.foyer_fiscal('nb_pac', period.last_year) > 0)
        aah_reduction = ((aah_date_debut_hospitalisation <= m_2) + (aah_date_debut_incarceration <= m_2)) * not_(pers_charge)

        return where(aah_reduction, aah_base * aah_parameters.pourcentage_aah.prison_hospitalisation, aah_base)
        # montant_max_aah = law.prestations_etat_de_sante.invalidite.aah.montant
        # est-ce cela, ou plutôt where(aah_reduction, min_(aah_base, aah_parameters.pourcentage_aah.prison_hospitalisation * montant_max_aah), aah_base)
        # ce qui expliquerait la phrase : L'intéressé ne peut recevoir une allocation plus élevée que celle qu'il percevrait s'il n'était pas hospitalisé, placé dans une maison d'accueil spécialisée ou incarcéré.
        # TODO: exemption de baisse également si paiement d'un forfait journalier (lors de l'hospitalisation), et si le conjoint ne travaille pas pour une raison reconnue valable
        # jusqu'en 2005, un taux différent selon si marié ou pas (https://www.legifrance.gouv.fr/codes/article_lc/LEGIARTI000031828843/2016-01-01/)


class eligibilite_caah(Variable):
    entity = Individu
    value_type = float
    label = "Eligibilité aux compléments à l'aah"
    reference = ['https://www.legifrance.gouv.fr/codes/article_lc/LEGIARTI000039802699']

    definition_period = MONTH
    set_input = set_input_dispatch_by_period

    def formula_2015_07_01(individu, period, parameters):
        annee_precedente = Period(('year', period.start, 1)).offset(-1)
        prestations = parameters(period).prestations_sociales
        taux_incapacite_min = prestations.prestations_etat_de_sante.invalidite.aah.taux_capacite.taux_incapacite
        aah = individu('aah', period)
        asi_eligibilite = individu('asi_eligibilite', period)
        asi = individu('asi', period)  # montant asi de la famille
        benef_asi = (asi_eligibilite * (asi > 0))
        taux_incapacite = individu('taux_incapacite', period)
        locataire_foyer = (individu.menage('statut_occupation_logement', period) == TypesStatutOccupationLogement.locataire_foyer)
        logement_independant = (individu.has_role(Menage.PERSONNE_DE_REFERENCE) + individu.has_role(Menage.CONJOINT)) * not_(locataire_foyer)

        activite_12_mois = individu('salaire_imposable', annee_precedente, options = [ADD]) + individu('rpns_imposables', annee_precedente)  # substitution à vérifier

        return (
            (taux_incapacite >= taux_incapacite_min)
            * ((aah > 0) | (benef_asi > 0))
            * logement_independant
            * (activite_12_mois == 0)
            )


class caah(Variable):
    calculate_output = calculate_output_add
    value_type = float
    label = "Complément d'allocation adulte handicapé (mensualisé)"
    entity = Individu
    set_input = set_input_divide_by_period
    definition_period = MONTH

    def formula_2019_12_01(individu, period, parameters):
        eligibilite_caah = individu('eligibilite_caah', period)
        mva = individu('mva', period)
        return mva * eligibilite_caah

    def formula_2015_07_01(individu, period, parameters):
        eligibilite_caah = individu('eligibilite_caah', period)
        complement_ressources_aah = individu('complement_ressources_aah', period)
        mva = individu('mva', period)
        return max_(complement_ressources_aah, mva) * eligibilite_caah

    def formula_2005_07_01(individu, period, parameters):
        invalidite = parameters(period).prestations_sociales.prestations_etat_de_sante.invalidite
        annee_precedente = Period(('year', period.start, 1)).offset(-1)
        activite_12_mois = individu('salaire_imposable', annee_precedente, options = [ADD]) + individu('rpns_imposables', annee_precedente)

        garantie_ressources = invalidite.caah.garantie_ressources
        aah_montant = invalidite.aah.montant

        aah = individu('aah', period)
        asi_eligibilite = individu('asi_eligibilite', period)
        asi = individu('asi', period)
        benef_asi = (asi_eligibilite * (asi > 0))

        # montant allocs logement de la famille
        al = individu.famille('aide_logement_montant', period)
        taux_incapacite = individu('taux_incapacite', period)
        # taux_capacite = individu('taux_capacite_travail', period) dans la législation, mais moins usité que le taux d'incapacité

        locataire_foyer = (individu.menage('statut_occupation_logement', period) == TypesStatutOccupationLogement.locataire_foyer)
        logement_independant = (individu.has_role(Menage.PERSONNE_DE_REFERENCE) + individu.has_role(Menage.CONJOINT)) * not_(locataire_foyer)
        incapacite = (taux_incapacite >= invalidite.aah.taux_capacite.taux_incapacite)

        elig_cpl = ((aah > 0) | (benef_asi > 0)) * incapacite * (activite_12_mois == 0) * logement_independant  # * non_capacite
        # TODO: revenus professionnels ?
        compl_ress = elig_cpl * max_(garantie_ressources - aah_montant, 0)

        elig_mva = (al > 0) * ((aah > 0) | (benef_asi > 0)) * incapacite * (activite_12_mois == 0) * logement_independant  # * non_capacite

        mva = invalidite.caah.majoration_vie_autonome * elig_mva

        return max_(compl_ress, mva)

    def formula_1994_07_01(individu, period, parameters):
        prestations_etat_de_sante = parameters(period).prestations_sociales.prestations_etat_de_sante
        cpltx = prestations_etat_de_sante.invalidite.caah.taux_montant_complement_ressources
        aah_montant = prestations_etat_de_sante.invalidite.aah.montant
        aah = individu('aah', period)
        asi_eligibilite = individu('asi_eligibilite', period)
        asi = individu('asi', period)
        benef_asi = (asi_eligibilite * (asi > 0))
        al = individu.famille('aide_logement_montant', period)
        taux_incapacite = individu('taux_incapacite', period)
        locataire_foyer = (individu.menage('statut_occupation_logement', period) == TypesStatutOccupationLogement.locataire_foyer)
        elig_ancien_caah = (al > 0) * ((aah > 0) | (benef_asi > 0)) * (taux_incapacite >= prestations_etat_de_sante.invalidite.aah.taux_capacite.taux_incapacite) * not_(locataire_foyer)
        ancien_caah = cpltx * aah_montant * elig_ancien_caah
        # En fait le taux cpltx perdure jusqu'en 2008 officiellement, la différence garantie-ressource et aah restant cependant constante égale à la valeur du complément d'allocation, 179,31
        return ancien_caah


class complement_ressources_aah(Variable):
    entity = Individu
    value_type = float
    label = 'Le complément de ressources'
    reference = 'https://www.legifrance.gouv.fr/affichCodeArticle.do?cidTexte=LEGITEXT000006073189&idArticle=LEGIARTI000006745305&dateTexte=&categorieLien=cid'
    definition_period = MONTH
    set_input = set_input_divide_by_period
    end = '2019-11-30'

    def formula_2015_07_01(individu, period, parameters):
        prestations = parameters(period).prestations_sociales
        garantie_ressources = prestations.prestations_etat_de_sante.invalidite.caah.garantie_ressources
        aah_montant = prestations.prestations_etat_de_sante.invalidite.aah.montant
        taux_capacite_travail_max = prestations.prestations_etat_de_sante.invalidite.aah.taux_capacite.taux_capacite_travail
        taux_capacite_travail = individu('taux_capacite_travail', period)

        return (taux_capacite_travail < taux_capacite_travail_max) * max_(garantie_ressources - aah_montant, 0)


class mva(Variable):
    entity = Individu
    value_type = float
    label = 'Majoration pour la vie autonome'
    reference = 'https://www.legifrance.gouv.fr/affichCodeArticle.do;jsessionid=6E5B97C7E6C7E06666BCFFA11871E70B.tplgfr43s_2?idArticle=LEGIARTI000006745350&cidTexte=LEGITEXT000006073189&dateTexte=20190124'
    definition_period = MONTH
    set_input = set_input_divide_by_period

    def formula_2015_07_01(individu, period, parameters):
        prestations = parameters(period).prestations_sociales
        al = individu.famille('aide_logement_montant', period)  # montant allocs logement de la famille
        mva_montant = prestations.prestations_etat_de_sante.invalidite.caah.majoration_vie_autonome

        return mva_montant * (al > 0)


class pch(Variable):  # inutilisée pour l'instant
    entity = Individu
    value_type = float
    label = 'Prestation de compensation du handicap'
    definition_period = MONTH
    set_input = set_input_dispatch_by_period
